;====================
; DXBCN project
; Controller of HAM radio beacon
; (C) Pavel Vachal,OK1DX 2012
; version 1.0 AtTiny26
; compiled in AVR Studio version 4
; 
; HAMWARE - free for Amateur radio community
; commercial use only with author's approval
;====================
; history: initial release 1.0

.NOLIST

;*********  FUNCTION LIST  *********
; LOOP	- main program loop, started after initialisation
; AD9850 - sends frequency value to DDS chip
; RD_EE - reads one byte from EEPROM
; STXT - sets CW text pointer to begin of selected text

;******** INTERRUPT HANDLERS *******
; RESET - entry point
; TIM0_OVF - timer 0 overflow: handling of CW timing

;******** FUSES ********************
; ATTiny26 needs to program fuses to operates correctly with external Xtal resonator 4 MHz
; lower fuses byte PLLCK=1, CKOPT=1, SUT1=1, SUT0=1, CKSEL3=1, CKSEL2=1, CKSEL1=1, CKSEL0=1 (PB4,PB5 XTAL) = 11111111=0xFF
; higher fuses byte = 11101111 = 0xEF (enable serial programming, PB7 is RESET), default value
; fuses programming must be done separately (not performed during code programming)!!!!

.include	"tn26def.inc"
.LIST

; port initialisation
; DDR bit 1=output, 0=input
; PORT bit when input 1=pullup 0=tri state
; Xtal freq = 5 MHz (not critical)
; port A:  PA0-3 spare, PA4 morse keying output, PA5-7 signals to DDS
.EQU	_DDRA	=0xF0	; PA0-PA4 input pullup, PA4-7 outputs
.EQU	_PORTA	=0x0F
; port B: PB0,1 text selector, PB3 spare, PB4,5 Xtal, PB6 spare, PB7=RESET
.EQU	_DDRB	=0x0	; PB0-7 inputs pullup
.EQU	_PORTB	=0xFF

; port A bits
.EQU CW_OUT = 4		; bit number
.EQU DDS_DATA = 5
.EQU DDS_CLK = 6
.EQU DDS_FR_UP = 7

; static RAM registers
.EQU	CWPTR = 0x60	; EEPROM CW text pointer
.EQU	CWCHAR = 0x61	; character being sent
.EQU	FREQ1 = 0x62	; freq + control, key up / 4 bytes
.EQU	FREQ2 = 0x66	; freq + control, key down / 4 bytes

; registers
.DEF	CWSPD   =R1	; speed constant from EEPROM
.DEF	CWCLK	=R2	; speed clock counter (up to CWSPD)
.DEF	ISREG	=R15	; temp SREG storage during processing of INT handler
.DEF	TMP1	=R16	; general temp regs used in main loop
.DEF	TMP2	=R17
.DEF	TMP3	=R18 
.DEF	CWREG	=R23	; bit 7 set when T0 timer expires, bit 6 elem being sent, 0-3 bits counters

; X register (r26 LOW BYTE, r27 HIGH BYTE - always zero) used as pointer to freq data

;********* MORSE CODE TABLE ********
.EQU	_A	=	0b110
.EQU	_B	=	0b10001
.EQU	_C	=	0b10101
.EQU	_D	=	0b1001
.EQU	_E	=	0b10
.EQU	_F	=	0b10100
.EQU	_G	=	0b1011
.EQU	_H	=	0b10000
.EQU	_I	=	0b100
.EQU	_J	=	0b11110
.EQU	_K	=	0b1101
.EQU	_L	=	0b10010
.EQU	_M	=	0b111
.EQU	_N	=	0b101
.EQU	_O	=	0b1111
.EQU	_P	=	0b10110
.EQU	_Q	=	0b11011
.EQU	_R	=	0b1010
.EQU	_S	=	0b1000
.EQU	_T	=	0b11
.EQU	_U	=	0b1100
.EQU	_V	=	0b11000
.EQU	_W	=	0b1110
.EQU	_X	=	0b11001
.EQU	_Y	=	0b11101
.EQU	_Z	=	0b10011
.EQU	_0	=	0b111111
.EQU	_1	=	0b111110
.EQU	_2	=	0b111100
.EQU	_3	=	0b111000
.EQU	_4	=	0b110000
.EQU	_5	=	0b100000
.EQU	_6	=	0b100001
.EQU	_7	=	0b100011
.EQU	_8	=	0b100111
.EQU	_9	=	0b101111
.EQU	_SL	=	0b101001	; /
.EQU	_AR	=	0b101010
.EQU	_SK	=	0b1101000
.EQU	_PT	=	0b1101010	; .
.EQU	_CM	=	0b1110011	; ,
.EQU	_EQ	=	0b110001	; =

.EQU	_SPC	=	1		; space
.EQU	_TEND	=	0	; end of text

.CSEG
.ORG	0

; interrupt vector table	
	rjmp RESET	;1 Reset Handler
	reti 		;2 INT0 Handler
	reti		;3 I/O Pins Handler
	reti		;4 TIMER1_COMP 1A
	reti		;5 TIMER1_COMP 1B
	reti		;6 TIMER1 overflow Handler
	rjmp TIM0_OVF	;7 TIMER0 overflow Handler
	reti		;8 USI start
	reti		;9 USI overflow
	reti		;10 EEPROM Ready Handler
	reti		;11 Analog Comparator Handler
	reti		;12 ADC Conversion Complete Interrupt Handler

RESET:
	ldi	TMP1,0xDF	; initiate stack pointer
	out	SP,TMP1

	ldi	TMP1,_DDRA	; port mode ini
	out	DDRA,TMP1
	ldi	TMP1,_DDRB
	out	DDRB,TMP1
	ldi	TMP1,_PORTA
	out	PORTA,TMP1
	ldi	TMP1,_PORTB
	out	PORTB,TMP1

; pulse on FR_UP / activate serial data transfer mode
	sbi	PORTA,DDS_FR_UP	; UPDATE FREQ pulse
	ldi	TMP2,6
RESET1:	dec	TMP2	; delay 4 microsec
	brne	RESET1	
	cbi	PORTA,DDS_FR_UP

; CPU register INI
	clr	CWREG
	rcall	STXT		; select CW text to be TXed
	ldi	TMP1,1
	sts	CWCHAR,TMP1

; read parameters from EEPROM
	ldi	TMP1,ECWSPD
	rcall	RD_EE
	mov	CWSPD, TMP1	; CW speed constant
	mov	CWCLK, TMP1

	clr	r27			; two FSK frequencies
	ldi	r26,FREQ1
	ldi	TMP3, 8		; 8 bytes total
	ldi	TMP2, EFREQ1	; pointer to EEPROM
PE1:	mov	TMP1, TMP2
	rcall	RD_EE
	st	X+,TMP1		; store to SRAM / FREQ1 and FREQ2
	inc	TMP2
	dec	TMP3
	brne	PE1	
	

; TIMER 0 initialisation
	ldi	TMP1,3
	out	TCCR0, TMP1	; T0 8bit internal/32
	ldi	TMP1,2
	out	TIMSK, TMP1		; enable TOIE0 interrupt only
	sei			; enable global interrupt

; ****** main program loop / performs most of the work
LOOP:

; CW message transmission
	sbrs	CWREG,7			; CW time expired
	rjmp	LOOP2
	cbr	CWREG,0x80		; clear the flag
	mov	TMP1,CWREG
	andi	TMP1, 0x0f		; get counter
	breq	CW1			; counter empty
	dec	CWREG
	rjmp 	LOOP2			; not yet empty, further wait
	
CW1:	sbrs	CWREG,6			; ? element over, insert 1 space
	rjmp	CW2
	ldi	CWREG,0x00		; KEY DWN after dot or dash elem = 0, length = 1
	cbi	PORTA,CW_OUT		; KEY = up
	clr	r27
	ldi	r26,FREQ1
	rcall	AD9850			; set DDS to FREQ1
	rjmp	LOOP2
	
CW2:	lds	TMP1,CWCHAR
	cpi	TMP1,1
	breq	CW3			; ? send space after char
	cpi	TMP1,0
	breq	CW4			; ? take next char
	
	lsr	TMP1
	brcs	CW5
	ldi	CWREG,0x40		; send DOT , length = 1
	rjmp	CW6
CW5:	ldi	CWREG,0x42		; send DASH , length = 3
CW6:	sbi	PORTA,CW_OUT		; KEY = down
	sts	CWCHAR,TMP1
	clr	r27
	ldi	r26,FREQ2
	rcall	AD9850			; set DDS to FREQ2
	rjmp	LOOP2
	
CW3:	ldi	CWREG,0x40		; SPACE AFTER CHAR elem = true, length = 1
	clr	TMP1
	sts	CWCHAR,TMP1
	rjmp	LOOP2
	
CW4:	clr	TMP2
	lds	TMP1,CWPTR
	mov	TMP3,TMP1
	rcall	RD_EE			; read one char from EEPROM
	inc	TMP3
	sts	CWPTR,TMP3
	sts	CWCHAR,TMP1
	cpi	TMP1,_TEND
	brne	CW2			; ? end of text
	rcall	STXT		; select start of CW text to be TXed
	rjmp	CW4			; read again from begining		
LOOP2:	nop			; end of CW message transmission

	rjmp	LOOP	;back in main loop

;******* SEND DATA to AD9850 **********
; data (4 bytes) taken from position in SRAM specified by X register
; LSB sent first, data followed by control byte
; takes 160 microsec

AD9850:
	cbi	PORTA,DDS_DATA
	cbi	PORTA,DDS_CLK
	cbi	PORTA,DDS_FR_UP	; all signals to low
	ldi	TMP3,4		; 4 bytes to send
AD3:	ld	TMP1,X+
	rcall	AD_BYTE
	dec	TMP3
	brne	AD3
	clr	TMP1
	rcall	AD_BYTE		; send control byte (phase=0, power on, control = 0)
	sbi	PORTA,DDS_FR_UP	; UPDATE FREQ pulse
	ldi	TMP2,6
AD1:	dec	TMP2	; delay 4 microsec
	brne	AD1
	cbi	PORTA,DDS_FR_UP
	ldi	TMP2,6
AD2:	dec	TMP2	; delay 4 microsec
	brne	AD2
	cbi	PORTA,DDS_DATA	; SERLOAD low
	ret
	
;***** AD_BYTE ******		; sends TMP1 (one byte) to AD9850, LSB first (used by AD9850 procedure)
AD_BYTE:	ldi	TMP2,8
AD_B1:	lsr	TMP1
	brcc	AD_B2
	sbi	PORTA,DDS_DATA	; bit=1, SERLOAD=1
	rjmp	AD_B3
AD_B2:	cbi	PORTA,DDS_DATA	; bit=0, SERLOAD=0
AD_B3:	nop
	nop
	nop
	nop
	sbi	PORTA,DDS_CLK	; W_CLK pulse 500 ns
	nop
	nop
	nop
	nop
	cbi	PORTA,DDS_CLK
	nop
	nop
	nop
	nop
	dec	TMP2
	brne	AD_B1
	ret

;******** PROCEDURE sets CW text pointer to begin of selected text ******
STXT:
	in	TMP1, PINB
	andi	TMP1, 0x3	; 4 texts specified by 2 input jumpers (text 0-3)
	ldi	TMP2, CWTXT
	add	TMP1,TMP2	; address in EEPROM containing pointers to text
	rcall	RD_EE		; read address of selected text from EEPROM
	sts	CWPTR,TMP1
	ret

;******** PROCEDURE reads one byte from EEPROM
RD_EE:	; input: address in TMP1; output: byte in TMP1
	out	EEAR,TMP1
	sbi	EECR,EERE
RDEE1:	sbic	EECR,EERE
	rjmp	RDEE1
	in	TMP1,EEDR	; text position byte
	ret

; ********* TIMER 0 overflow HANDLER ***********
; sets a flag indicating it's time to process next morse element
TIM0_OVF:
		in	ISREG,SREG
		dec	CWCLK
		brne	TIM0_1
		mov	CWCLK,CWSPD	; set new speed value
		sbr	CWREG,0x80	; set flag - time expired
TIM0_1:	out	SREG,ISREG
		reti

;******** EEPROM CONTENTS DEFINITION
.ESEG
.ORG	0x0

; frequency calculation - 32bit value = freq * 2^32 / 125000000
; example 50.050 MHz -> 0x66809d49, 50.0505 MHz -> 0x6680E065
EFREQ1:	.DB		0x49, 0x9d, 0x80, 0x66	; frequency constant, LSB first; key up
EFREQ2:	.DB		0x65, 0xE0, 0x80, 0x66	; frequency constant, LSB first; key down
ECWSPD:	.DB		20, 0
CWTXT:	.DB		CWTXT3,CWTXT2,CWTXT1,CWTXT0
CWTXT0:	.DB		_O, _K, _1, _D, _X, _SL, _B, _SPC, _J, _N, _6, _9, _S, _D, _SPC, _SPC, _SPC, _TEND
CWTXT1:	.DB		_O, _K, _1, _D, _X, _SL, _B, _SPC, _J, _N, _6, _9, _N, _O, _SPC, _SPC, _SPC, _TEND
CWTXT2:	.DB		_O, _K, _1, _D, _X, _SPC, _B, _E, _A, _C, _O, _N, _SPC, _J, _N, _6, _9, _S, _D, _SPC, _SPC, _SPC, _TEND
CWTXT3:	.DB		_O, _K, _0, _B, _M, _SL, _B, _SPC, _J, _N, _6, _9, _S, _D, _SPC, _SPC, _SPC, _TEND
